// 倒立振子、加速度/ジャイロセンサー(MPU-6050)版
// 2016/3/4

#include <Wire.h>

// モータードライバの制御定数。
#define MOTOR_L_1 3 // L298 INPUT 1
#define MOTOR_L_2 5 // L298 INPUT 2

#define MOTOR_R_1 3 // L298 INPUT 3
#define MOTOR_R_2 5 // L298 INPUT 4

#define V_MIN 100 // モーター駆動電圧(0～255)の最低値。これより低いと回らない。
#define V_MAX 200 // モーター駆動電圧(0～255)の最高値。

// 加速度/ジャイロセンサーの制御定数。
#define MPU6050_ADDR         0x68 // MPU-6050 device address
#define MPU6050_SMPLRT_DIV   0x19 // MPU-6050 register address
#define MPU6050_CONFIG       0x1a
#define MPU6050_GYRO_CONFIG  0x1b
#define MPU6050_ACCEL_CONFIG 0x1c
#define MPU6050_ACCEL_XOUT_H 0x3b
#define MPU6050_ACCEL_XOUT_L 0x3c
#define MPU6050_ACCEL_YOUT_H 0x3d
#define MPU6050_ACCEL_YOUT_L 0x3e
#define MPU6050_ACCEL_ZOUT_H 0x3f
#define MPU6050_ACCEL_ZOUT_L 0x40
#define MPU6050_GYRO_XOUT_H  0x43
#define MPU6050_GYRO_XOUT_L  0x44
#define MPU6050_GYRO_YOUT_H  0x45
#define MPU6050_GYRO_YOUT_L  0x46
#define MPU6050_GYRO_ZOUT_H  0x47
#define MPU6050_GYRO_ZOUT_L  0x48
#define MPU6050_PWR_MGMT_1   0x6b
#define MPU6050_WHO_AM_I     0x75

float power,pitch;
float kpitch=0.015;
int dur=180;

// 倒立振子の制御変数
float ax,ay,az;
float degRoll,degPitch;
float gx,gy,gz;
float dpsX,dpsY,dpsZ;

// 加速度/ジャイロセンサーへのコマンド送信。
void writeMPU6050(byte reg, byte data) {
  Wire.beginTransmission(MPU6050_ADDR);
  Wire.write(reg);
  Wire.write(data);
  Wire.endTransmission();
}

// 加速度/ジャイロセンサーからのデータ読み込み。
byte readMPU6050(byte reg) {
  Wire.beginTransmission(MPU6050_ADDR);
  Wire.write(reg);
  Wire.endTransmission(false);
  Wire.requestFrom(MPU6050_ADDR, 1/*length*/, false);
  byte data =  Wire.read();
  Wire.endTransmission(true);
  return data;
}

void setup() {
  Serial.begin(57600);

  // モータードライバの初期化。
  pinMode(MOTOR_L_1, OUTPUT);
  pinMode(MOTOR_L_2, OUTPUT);

  pinMode(MOTOR_R_1, OUTPUT);
  pinMode(MOTOR_R_2, OUTPUT);

  // 加速度/ジャイロセンサーの初期化。
  Wire.begin();
  if (readMPU6050(MPU6050_WHO_AM_I) != 0x68) {
    Serial.println("\nWHO_AM_I error.");
    while (true) ;
  }
  writeMPU6050(MPU6050_SMPLRT_DIV, 0x07);   // sample rate: 8kHz/(7+1) = 1kHz
  writeMPU6050(MPU6050_CONFIG, 0x00);       // disable DLPF, gyro output rate = 8kHz
  writeMPU6050(MPU6050_GYRO_CONFIG, 0x00);  // gyro range: ±250dps
  writeMPU6050(MPU6050_ACCEL_CONFIG, 0x00); // accel range: ±2g
  writeMPU6050(MPU6050_PWR_MGMT_1, 0x01);   // disable sleep mode, PLL with X gyro
  delay(2000);
}

void loop() {
  pitch=getSensor();
  power=max(min(kpitch*pitch,1),-1);
  Serial.print(power*50);
  Serial.print(" ");
  // power : -1～1
  float powerL = power;
  float powerR = power;
  
  // 方向転換させる場合はここでpowerL, Rに差分を与える。

  // powerをモーター駆動電圧に変換。0～1 → V_MIN～V_MAX
  int voltL = (int)((V_MAX - V_MIN) * abs(powerL) + V_MIN);
  int voltR = (int)((V_MAX - V_MIN) * abs(powerR) + V_MIN);
  Serial.println(voltL);
  // モーター駆動
  if (power>0.1){
    analogWrite(MOTOR_L_1, voltL);
    analogWrite(MOTOR_L_2,0);
    analogWrite(MOTOR_R_1, voltR);
    analogWrite(MOTOR_R_2,0);
    delay(dur);
  } else if(power<-0.1){
    analogWrite(MOTOR_L_1, 0);
    analogWrite(MOTOR_L_2, voltL);
    analogWrite(MOTOR_R_1, 0);
    analogWrite(MOTOR_R_2, voltR);
    delay(dur);
  }

  // 倒れたらモーター停止。
  if (abs(power)>230 ) {
    analogWrite(MOTOR_L_1, 0);
    analogWrite(MOTOR_R_2, 0);
    Serial.print("**** stop ");
    Serial.print(degPitch);
    while (1) delay(1000);
  }
}

float getSensor(){
  // 重力加速度から角度を求める。
  ax = (readMPU6050(MPU6050_ACCEL_XOUT_H) << 8) | readMPU6050(MPU6050_ACCEL_XOUT_L);
  ay = (readMPU6050(MPU6050_ACCEL_YOUT_H) << 8) | readMPU6050(MPU6050_ACCEL_YOUT_L);
  az = (readMPU6050(MPU6050_ACCEL_ZOUT_H) << 8) | readMPU6050(MPU6050_ACCEL_ZOUT_L);
  degRoll  = atan2(ay, az) * RAD_TO_DEG;
  degPitch = atan(-ax / sqrt(ay * ay + az * az)) * RAD_TO_DEG;

  // ジャイロで角速度を求める。
  gx = (readMPU6050(MPU6050_GYRO_XOUT_H) << 8) | readMPU6050(MPU6050_GYRO_XOUT_L);
  gy = (readMPU6050(MPU6050_GYRO_YOUT_H) << 8) | readMPU6050(MPU6050_GYRO_YOUT_L);
  gz = (readMPU6050(MPU6050_GYRO_ZOUT_H) << 8) | readMPU6050(MPU6050_GYRO_ZOUT_L);
  dpsX = gx / 131.0; // LSB sensitivity: 131 LSB/dps @ ±250dps
  dpsY = gy / 131.0;
  dpsZ = gz / 131.0;
  
  return degPitch;
}


